/*  ti397.c: Routines for handling the FDD3000 interface 
    (based on the +D routines)

   Copyright (c) 1999-2016 Stuart Brady, Fredrick Meunier, Philip Kendall,
   Dmitry Sanarin, Darren Salt, 2014 Gergely Szasz, Sławomir Szczyrba

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   Author contact information:

   Philip: philip-fuse@shadowmagic.org.uk

   Stuart: stuart.brady@gmail.com

*/

#include <config.h>

#include <libspectrum.h>

#include <string.h>

#include "compat.h"
#include "debugger/debugger.h"
#include "infrastructure/startup_manager.h"
#include "machine.h"
#include "module.h"
#include "ti397.h"
#include "fdd3000.h"
#include "peripherals/printer.h"
#include "settings.h"
#include "ui/ui.h"
#include "ui/uimedia.h"
#include "unittests/unittests.h"
#include "utils.h"
#include "wd_fdc.h"
#include "options.h"    /* needed for get combo options */
#include "z80/z80.h"

/* 8KB ROM */
#define FDD3000_ROM_SIZE 0x1000
/* 8KB RAM */
#define FDD3000_RAM_SIZE 0x2000

static int fdd3000_rom_memory_source, fdd3000_ram_memory_source;

/* Two memory chunks accessible by the Z80 when /ROMCS is low */
static memory_page fdd3000_memory_map_romcs_rom[ MEMORY_PAGES_IN_8K ];
static memory_page fdd3000_memory_map_romcs_ram[ MEMORY_PAGES_IN_8K ];

int fdd3000_available = 0;
int fdd3000_active = 0;

static int current_drv, side_b, head_up, double_density;

wd_fdc *fdd3000_fdc;
static fdd_t fdd3000_drives[ FDD3000_NUM_DRIVES ];
static ui_media_drive_info_t fdd3000_ui_drives[ FDD3000_NUM_DRIVES ];

static libspectrum_byte fdd3000_ram[FDD3000_RAM_SIZE];

static void fdd3000_reset( int hard_reset );
static void fdd3000_memory_map( void );
static void fdd3000_activate( void );



static void
fdd3000_io_write( libspectrum_word port GCC_UNUSED, libspectrum_byte b )
{

  if( ( port & 0xff ) == 0xef ) {
    wrTI2FDD( b );
  }
}

static libspectrum_byte
fdd3000_io_read( libspectrum_word port GCC_UNUSED, libspectrum_byte *attached )
{
  libspectrum_byte ret = 0;
  if( ( port & 0xff ) == 0xef ) {
    *attached = 0xff;
    ret = rdFDD2TI();
  }
  return ret;
}

static const periph_port_t fdd3000_ports[] = {
  { 0x00ff, 0x00ef, fdd3000_io_read, fdd3000_io_write },
  { 0, 0, NULL, NULL }
};

static module_info_t fdd3000_module_info = {

  /* .reset = */ fdd3000_reset,
  /* .romcs = */ fdd3000_memory_map,
  /* .snapshot_enabled = */ 0,
  /* .snapshot_from = */ 0,
  /* .snapshot_to = */ 0

};

static const periph_t fdd3000_periph = {
  /* .option = */ &settings_current.fdd3000,
  /* .ports = */ fdd3000_ports,
  /* .hard_reset = */ 1,
  /* .activate = */ fdd3000_activate
};

/* Debugger events */
static const char *const event_type_string = "fdd3000";
static int page_event, unpage_event;

void
fdd3000_page( void )
{
    debug_print("FDD3 page\n");    
  fdd3000_active = 1;
  machine_current->ram.romcs = 1;
  machine_current->memory_map();
  debugger_event( page_event );
}

void
fdd3000_unpage( void )
{
    debug_print("FDD3 unpage\n");
  fdd3000_active = 0;
  machine_current->ram.romcs = 0;
  machine_current->memory_map();
  debugger_event( unpage_event );
}

void
dump_memory_page( memory_page mp )
{
  fprintf( stderr, "page ptr  : %p\n", mp.page );
  fprintf( stderr, "writable  : %d\n", mp.writable );
  fprintf( stderr, "contended : %d\n", mp.contended );
  fprintf( stderr, "source    : %d\n", mp.source );
  fprintf( stderr, "snapshot  : %d\n", mp.save_to_snapshot );
  fprintf( stderr, "page num  : %d\n", mp.page_num );
  fprintf( stderr, "offset    : x%04x\n\n", mp.offset );
}


static void
fdd3000_memory_map( void )
{
  if( !fdd3000_active ) return;
  memory_map_romcs_8k( 0x0000, fdd3000_memory_map_romcs_rom );
  memory_map_romcs_8k( 0x2000, fdd3000_memory_map_romcs_ram );
}

static void
fdd3000_activate( void )
{
  debug_print("FDD3 activate\n");;
}

static void
fdd3000_set_datarq( struct wd_fdc *f )
{
  fddSetDRQ();
}
static void
fdd3000_reset_datarq( struct wd_fdc *f )
{
  fddClrDRQ();
}
static void
fdd3000_set_intrq( struct wd_fdc *f )
{
  fddSetFDCInt( 0xff );
}
static void
fdd3000_reset_intrq( struct wd_fdc *f )
{
  fddClrFDCInt();
}

static int
fdd3000_init( void *context )
{
  int i;
  fdd_t *d;

    debug_print("FDD3 init\n");

  fdd3000_fdc = wd_fdc_alloc_fdc( WD1772, 0, WD_FLAG_NONE | WD_FLAG_FDD3000 );

  for( i = 0; i < FDD3000_NUM_DRIVES; i++ ) {
    d = &fdd3000_drives[ i ];
    fdd_init( d, FDD_SHUGART, NULL, 0 ); /* drive geometry 'autodetect' */
    d->disk.flag = DISK_FLAG_NONE;/*|DISK_FLAG_PLUS3_CPC; */
  }

  fdd3000_fdc->current_drive = &fdd3000_drives[ 0 ];
  fdd_select( &fdd3000_drives[ 0 ], 1 );
  fdd3000_fdc->dden = 1;
  fdd3000_fdc->set_intrq = fdd3000_set_intrq;
  fdd3000_fdc->reset_intrq = fdd3000_reset_intrq;
  fdd3000_fdc->set_datarq = fdd3000_set_datarq;
  fdd3000_fdc->reset_datarq = fdd3000_reset_datarq;

  module_register( &fdd3000_module_info );

  fdd3000_rom_memory_source = memory_source_register( "FDD3000 ROM" );
  fdd3000_ram_memory_source = memory_source_register( "FDD3000 RAM" );

  for( i = 0; i < MEMORY_PAGES_IN_8K; i++ )
    fdd3000_memory_map_romcs_rom[i].source = fdd3000_rom_memory_source;
  for( i = 0; i < MEMORY_PAGES_IN_8K; i++ )
    fdd3000_memory_map_romcs_ram[i].source = fdd3000_ram_memory_source;

  periph_register( PERIPH_TYPE_FDD3000, &fdd3000_periph );

  for( i = 0; i < FDD3000_NUM_DRIVES; i++ ) {
    fdd3000_ui_drives[ i ].fdd = &fdd3000_drives[ i ];
    ui_media_drive_register( &fdd3000_ui_drives[ i ] );
  }

  periph_register_paging_events( event_type_string, &page_event,
                                 &unpage_event );
  ui_menu_activate( UI_MENU_ITEM_MACHINE_FDD3000_RESET, 1 );
  current_drv = side_b = head_up = double_density = 0;
  wd_fdc_master_reset( fdd3000_fdc );
  fddReset();

  for( i = 0; i < FDD3000_NUM_DRIVES; i++ ) {
    ui_media_drive_update_menus( &fdd3000_ui_drives[ i ],
                                 UI_MEDIA_DRIVE_UPDATE_ALL );
  }

  fdd3000_fdc->current_drive = &fdd3000_drives[ 0 ];
  fdd_select( &fdd3000_drives[ 0 ], 0 );
  return 0;
}

static void
fdd3000_end( void )
{
    debug_print("FDD3 end\n");
  fdd3000_available = 0;
  libspectrum_free( fdd3000_fdc );
  ui_menu_activate( UI_MENU_ITEM_MACHINE_FDD3000_RESET, 0 );

}

void
fdd3000_register_startup( void )
{
    debug_print("FDD3 register startup\n");
  startup_manager_module dependencies[] = {
    STARTUP_MANAGER_MODULE_DEBUGGER,
    STARTUP_MANAGER_MODULE_MEMORY,
    STARTUP_MANAGER_MODULE_SETUID,
  };
  startup_manager_register( STARTUP_MANAGER_MODULE_FDD3000, dependencies,
                            ARRAY_SIZE( dependencies ), fdd3000_init, NULL,
                            fdd3000_end );
}

static void
fdd3000_reset( int hard_reset )
{
  int i;

    debug_print("FDD3 reset\n");

  fdd3000_active = 0;
  fdd3000_available = 0;

  if( !periph_is_active( PERIPH_TYPE_FDD3000 ) ) {
    ui_menu_activate( UI_MENU_ITEM_MACHINE_FDD3000_RESET, 0 );
    return;
  }
  if( machine_load_rom_bank( fdd3000_memory_map_romcs_rom, 0,
                             settings_current.rom_fdd3000,
                             settings_default.rom_fdd3000,
                             FDD3000_ROM_SIZE ) ) {
    settings_current.fdd3000 = 0;
    periph_activate_type( PERIPH_TYPE_FDD3000, 0 );
    return;
  }
  for( i = 0; i < MEMORY_PAGES_IN_4K; i++ ) {
    fdd3000_memory_map_romcs_rom[ i +
                                  MEMORY_PAGES_IN_4K ] =
      fdd3000_memory_map_romcs_rom[ i ];
  }

  machine_current->ram.romcs = 0;

  for( i = 0; i < MEMORY_PAGES_IN_8K; i++ ) {
    fdd3000_memory_map_romcs_ram[ i ].page = &fdd3000_ram[i * MEMORY_PAGE_SIZE];
    fdd3000_memory_map_romcs_ram[ i ].offset = i * MEMORY_PAGE_SIZE;
    fdd3000_memory_map_romcs_ram[ i ].writable = 1;
  }

  fdd3000_available = 1;

  if( hard_reset )
    memset( fdd3000_ram, 0, FDD3000_RAM_SIZE );

  machine_current->memory_map();
  ui_menu_activate( UI_MENU_ITEM_MACHINE_FDD3000_RESET, 1 );
}


fdd_t *
fdd3000_get_fdd( fdd3000_drive_number which )
{
  return &( fdd3000_drives[ which ] );
}

int
fdd3000_unittest( void )
{
  int r = 0;

  fdd3000_page();

  r += unittests_assert_4k_page( 0x0000, fdd3000_rom_memory_source, 0 );
  r += unittests_assert_2k_page( 0x2000, fdd3000_ram_memory_source, 0 );
  /* FIXME: should we add mirroring at 0x2800, 0x3000 and/or 0x3800? */
  r += unittests_assert_16k_ram_page( 0x4000, 5 );
  r += unittests_assert_16k_ram_page( 0x8000, 2 );
  r += unittests_assert_16k_ram_page( 0xc000, 0 );

  fdd3000_unpage();

  r += unittests_paging_test_48( 2 );

  return r;
}

static int
ui_drive_is_available( void )
{
  return fdd3000_available;
}

static const fdd_params_t *
ui_drive_get_params_1( void )
{
  return &fdd_params[ option_enumerate_diskoptions_drive_fdd3000a_type() + 1 ];
}

static const fdd_params_t *
ui_drive_get_params_2( void )
{
  return &fdd_params[ option_enumerate_diskoptions_drive_fdd3000b_type() ];
}

static const fdd_params_t *
ui_drive_get_params_3( void )
{
  return &fdd_params[ option_enumerate_diskoptions_drive_fdd3000c_type() ];
}

static const fdd_params_t *
ui_drive_get_params_4( void )
{
  return &fdd_params[ option_enumerate_diskoptions_drive_fdd3000d_type() ];
}

int
fdd3000_disk_insert( fdd3000_drive_number which, const char *filename,
                     int autoload )
{
  if( which >= FDD3000_NUM_DRIVES ) {
    ui_error( UI_ERROR_ERROR, "fdd3000_disk_insert: unknown drive %d",
              which );
    fuse_abort();
  }

  return ui_media_drive_insert( &fdd3000_ui_drives[ which ], filename,
                                autoload );
}


static ui_media_drive_info_t fdd3000_ui_drives[ FDD3000_NUM_DRIVES ] = {
  {
    /* .name = */ "FDD3000 Disk A",
    /* .controller_index = */ UI_MEDIA_CONTROLLER_FDD3000,
    /* .drive_index = */ FDD3000_DRIVE_A,
    /* .menu_item_parent = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000,
    /* .menu_item_top = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_A,
    /* .menu_item_eject = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_A_EJECT,
    /* .menu_item_flip = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_A_FLIP_SET,
    /* .menu_item_wp = */ 0,/*UI_MENU_ITEM_MEDIA_DISK_FDD3000_A_WP_SET, */
    /* .is_available = */ &ui_drive_is_available,
    /* .get_params = */ &ui_drive_get_params_1,
  },
  {
    /* .name = */ "FDD3000 Disk B",
    /* .controller_index = */ UI_MEDIA_CONTROLLER_FDD3000,
    /* .drive_index = */ FDD3000_DRIVE_B,
    /* .menu_item_parent = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000,
    /* .menu_item_top = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_B,
    /* .menu_item_eject = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_B_EJECT,
    /* .menu_item_flip = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_B_FLIP_SET,
    /* .menu_item_wp = */ 0,/*UI_MENU_ITEM_MEDIA_DISK_FDD3000_B_WP_SET, */
    /* .is_available = */ &ui_drive_is_available,
    /* .get_params = */ &ui_drive_get_params_2,
  },
  {
    /* .name = */ "FDD3000 Disk C",
    /* .controller_index = */ UI_MEDIA_CONTROLLER_FDD3000,
    /* .drive_index = */ FDD3000_DRIVE_C,
    /* .menu_item_parent = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000,
    /* .menu_item_top = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_C,
    /* .menu_item_eject = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_C_EJECT,
    /* .menu_item_flip = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_C_FLIP_SET,
    /* .menu_item_wp = */ 0,/*UI_MENU_ITEM_MEDIA_DISK_FDD3000_C_WP_SET, */
    /* .is_available = */ &ui_drive_is_available,
    /* .get_params = */ &ui_drive_get_params_3,
  },
  {
    /* .name = */ "FDD3000 Disk D",
    /* .controller_index = */ UI_MEDIA_CONTROLLER_FDD3000,
    /* .drive_index = */ FDD3000_DRIVE_D,
    /* .menu_item_parent = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000,
    /* .menu_item_top = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_D,
    /* .menu_item_eject = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_D_EJECT,
    /* .menu_item_flip = */ UI_MENU_ITEM_MEDIA_DISK_FDD3000_D_FLIP_SET,
    /* .menu_item_wp = */ 0,/*UI_MENU_ITEM_MEDIA_DISK_FDD3000_D_WP_SET, */
    /* .is_available = */ &ui_drive_is_available,
    /* .get_params = */ &ui_drive_get_params_4,
  },
};

/*
; bit 0 - /select drive 0
; bit 1 - /select drive 1
; bit 2 - /select drive 2
; bit 3 - /select drive 3
; bit 4 - select side A-/B
; bit 5 - double density /DDEN
; bit 6 - select /BOOTROM
; bit 7 - /head up ( 0 - głowica aktywna-leży na dyskietce, 1 - głowica nieaktywna-w powietrzu)
*/


void
setDriveCfg( uint8_t val )
{
  int drv = val & 0x0f;

  switch( drv ) {
  case FDD3_DRIVE_NONE: current_drv = 4; break;
  case FDD3_DRIVE_0: current_drv = 0; break;
  case FDD3_DRIVE_1: current_drv = 1; break;
  case FDD3_DRIVE_2: current_drv = 2; break;
  case FDD3_DRIVE_3: current_drv = 3; break;
  }
  if( val & FDD3_SIDE_B ) {
    side_b = 1;
  } else {
    side_b = 0;
  }
  if( val & FDD3_DD ) {
    double_density = 1;
  } else {
    double_density = 0;
  }
  if( val & FDD3_HEAD ) {
    head_up = 1;
  } else {
    head_up = 0;
  }

  if( current_drv < 4 ) {/* */
    fdd_motoron( fdd3000_fdc->current_drive, 1 );
    fdd_head_load( fdd3000_fdc->current_drive, 1 );
    fdd_select( fdd3000_fdc->current_drive, 0 );
    fdd3000_fdc->dden = double_density;

    fdd3000_fdc->current_drive = &fdd3000_drives[ current_drv ];
    fdd_select( fdd3000_fdc->current_drive, 1 );
    fdd_set_head( fdd3000_fdc->current_drive, side_b );
    fdd_head_load( fdd3000_fdc->current_drive, head_up );
    /* fdd_motoron(fdd3000_fdc->current_drive, 1);
       fdd3000_fdc->current_drive->unreadable=(current_drv==4);	// */
  } else {   /* */
    fdd_select( fdd3000_fdc->current_drive, 0 );       /* */
  }   /* */
}

void
wrCommandReg( uint8_t val )
{
  wd_fdc_cr_write( fdd3000_fdc, val );
}

void
wrTrackReg( uint8_t val )
{
  wd_fdc_tr_write( fdd3000_fdc, val );
}

void
wrSectorReg( uint8_t val )
{
  wd_fdc_sec_write( fdd3000_fdc, val );
}

void
wrDataReg( uint8_t val )
{
  wd_fdc_dr_write( fdd3000_fdc, val );
}

uint8_t
rdStatusReg( void )
{
  return wd_fdc_sr_read( fdd3000_fdc );
}

uint8_t
rdTrackReg( void )
{
  return wd_fdc_tr_read( fdd3000_fdc );
}

uint8_t
rdSectorReg( void )
{
  return wd_fdc_sec_read( fdd3000_fdc );
}

uint8_t
rdDataReg( void )
{
  return wd_fdc_dr_read( fdd3000_fdc );
}

